import std.string;

struct FFI {
    version (X86) {
    }
    else {
        static assert(false, "unsupported architecture");
    }

    void* fp;
    TypeInfo retType;
    TypeInfo[] argTypes;
    void* ret;
    void*[] args;
    bool isVargs;

    void setFp(void* p) {
        fp = p;
    }

    void addArg(T)(T* x) {
        argTypes ~= typeid(T);
        args ~= x;
    }

    void setRet(T)(T* x) {
        retType = typeid(T);
        ret = x;
    }

    void setIsVargs() {
        isVargs = true;
    }

    int call() {
        assert(fp);
        assert(args.length == argTypes.length);

        void* self = this;

        bool should_pop = false;

        for (int i = 0; i < args.length; i++) {
            int asize = argTypes[i].tsize;
            switch (asize) {
            case 1:
            case 2:
            case 4:
                should_pop = true;
                int v = *cast(int*)args[i];
                asm {
                    push v;
                }
                break;
            default:
                should_pop = false;
                int t = (asize+3) >> 2;
                while (t--) {
                    int v = (cast(int*)args[i])[t];
                    asm {
                        push v;
                    }
                }
                break;
            }
        }

        if (isVargs) {
            should_pop = false;
            TypeInfo_Tuple tuple = new TypeInfo_Tuple();
            tuple.elements = argTypes;
            asm {
                push tuple;
            }
        }

        int rsize = ret ? retType.tsize : 0;
        if (rsize > 2 && rsize != 4 && rsize != 8) {
            asm {
                mov EBX, self;
                mov EBX, ret[EBX];
                push EBX;
            }
            should_pop = true;
        }

        if (should_pop) {
            asm {
                mov EBX, self;
                mov EBX, fp[EBX];
                pop EAX;
                call EBX;
                push EDX;
                push EAX;
            }
        }
        else {
            asm {
                mov EBX, self;
                mov EBX, fp[EBX];
                call EBX;
                push EDX;
                push EAX;
            }
        }

        switch (rsize) {
        case 1:
        case 2:
        case 4:
            asm {
                pop EAX;
                mov EBX, self;
                mov EBX, ret[EBX];
                mov [EBX], EAX;
            }
            break;
        case 8:
            asm {
                pop EAX;
                pop EDX;
                mov ESI, self;
                mov ESI, ret[ESI];
                mov [ESI], EAX;
                mov 4[ESI], EDX;
            }
            break;
        default:
            break;
        }

        return 0;
    }
}

